// @TITLE "ActiveSet.cc: class ActiveSet"
//
//  This class implements the active set.  There is only one active
// set, which determines which VPs are active at any given time.  The
// ifp() macro, however, can temporarily restrict the active set to the
// subset where the condition is true, and then to where the condition is
// false.  These can be nested, so the active set actually depends on an
// active stack to manage a stack of active sets.
//   There are some methods of this class which are public, but only because
// they are used by the macros; the end user is not supposed to use these
// functions directly.
//   Protected member isActive() is called by the Collection functions to 
// decide whether a given VP is active.  We in turn are allowed to read
// the v[] component of the collection we use to represent the current set.
// 
// Internally, we use another class called ActiveStack, which is
// used to represent the stack of active states.
//
// David Kotz 1994

// $Id: ActiveSet.cc,v 2.16 95/12/17 00:08:03 dfk release Locker: dfk $

#include "dapple.h"

ActiveSet Active;		// the one and only, current ActiveSet.

// #define DEBUG		// prints some debugging info

// A stack is really a linked list of ActiveStack objects. 

class ActiveStack {
private:
    friend class ActiveSet;	// only ActiveSet may use this class
    ActiveStack(ActiveStack *n, booleanCollection *a, 
		booleanCollection *c, CONST int lp) {
	next = n; active = a; condition = c; loop = lp;
    }
    ~ActiveStack(void) {  }	// not MY job to empty the stack

    booleanCollection *active;    // copy of the active set at this level
    booleanCollection *condition; // copy of the condition to get to current
    int loop;			    // which clause are we in: then or else?
    ActiveStack *next;		    // pointer to next thing in stack
};


/////////////////////////////////////////////////////////
// @SUBTITLE "ActiveSet::ActiveSet"
//   Construct the active set.

ActiveSet::ActiveSet()
{
    current = NULL;		// outer scope is special; all active
    everywhere = false;		// only true for brief periods
    whichLoop = 0;		// re-initialized in restrict()

    enclosing = NULL;		// and there is no enclosing state
}

/////////////////////////////////////////////////////////
// @SUBTITLE "ActiveSet::~ActiveSet"
//   Destroy the active set object.  This should probably never be called.

ActiveSet::~ActiveSet()
{
    if (enclosing)		// shouldn't be happening
      delete enclosing;
}

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "ActiveSet::restrict: ifp support"
//   See ActiveSet.h for #define ifp().
// This function is called at the beginning of ifp().  We need to save 
// both the current active set and the condition, for future use by
// which below.  We do this by pushing the current active set
// and the current condition onto a stack, and then making a new
// active set and condition that are used for now.    We also need to save
// the current then/else flag, so we can make a new one; the current is
// pushed onto the same stack with the other information.

void
ActiveSet::restrict(CONST booleanCollection& condition)
{
    everywhere = true;		// so we can work with whole collections

#ifdef DEBUG
    if (current)
      cout << "Pushing active set: " << *current << endl;
    else
      cout << "Pushing active set: (OUTER)" << endl;
    cout << "pushing condition: " << condition << endl;
#endif
    
    // make a copy of the condition to save on the stack
    booleanCollection *conditionCopy = new booleanCollection(condition);
    INVARIANT2(conditionCopy != NULL, "(internal) ActiveSet found NULL conditionCopy");

    // push the current state onto the stack
    enclosing = new ActiveStack(enclosing, current, conditionCopy, whichLoop);
    INVARIANT2(enclosing != NULL, "(internal) no memory for new ActiveStack");

    whichLoop = 0;		// ie, the "then" clause

    everywhere = false;
}

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "ActiveSet::loop: ifp support"
//   See ActiveSet.h for #define ifp().
// We can be called in one of three states:
// whichLoop = 0:
//  	just before doing the THEN clause; yes, we should loop
// whichLoop = 1:
//  	just before doing the ELSE clause; yes, we should loop
// whichLoop = 2:
//  	just after doing the ELSE clause; no, we should not loop
//  	AND, we should pop the active stack, because ifp() is over.

boolean				// keep looping? true twice, false once
ActiveSet::loop(void)
{
    boolean status;		// do we keep looping? 

    everywhere = true;		// so we can manipulate the Active set

    switch (whichLoop) {
	case 0: {		// just before THEN clause
	    // set active set, based on the old one and the new condition
	    if (enclosing->active == NULL)
	      current = new booleanCollection(*(enclosing->condition));	// current = condition
	    else {
		current = new booleanCollection(*(enclosing->condition)); // current = condition
		OP_AND(*current, *current, *(enclosing->active) ); // current = current && active
	    }
	    status = true;
	    break;
	}
		
	case 1: {		// just before ELSE clause
	    // set active set, based on the old one and the new condition
	    if (enclosing->active == NULL)
	      OP_NOT(*current, *(enclosing->condition) ); // current = !condition
	    else {
		OP_NOT(*current, *(enclosing->condition));  // current = !condition
		OP_AND(*current, *current, *(enclosing->active) ); // current = current && active
	    }
	    status = true;
	    break;
	}

	case 2: {		// just after ELSE clause; time to quit
	    // pop back the old active set and whichLoop
	    //   (note the parameters to pop are reference variables)
	    ActiveStack *old = enclosing;
	    INVARIANT2(enclosing != NULL, 
		       "Illegal nesting of ifp(); perhaps you used 'break' inside ifp()?");
	    delete enclosing->condition; // get rid of the saved copy
	    delete current;		 // get rid of the current active set
	    current = enclosing->active; // restore active set
	    whichLoop = enclosing->loop; // restore whichLoop
	    enclosing = enclosing->next; // pop

	    delete old;		 // and delete the unneeded stack entry
	    
	    status = false;
	    break;
	}

	default: {		// huh?
	    // this shouldn't happen, but it could happen if someone
	    // executed "break;" inside of a "then" or "else" clause
	    INVARIANT2(enclosing != NULL, 
		       "Illegal nesting of ifp(); perhaps you used 'break' inside ifp()?");
	    break;
	}
    }
		
#ifdef DEBUG
    if (current)
      cout << "Active set now (whichLoop now " << whichLoop << "): (OUTER)" 
	<< endl;
    else
      cout << "Active set now (whichLoop now " << whichLoop << "): " 
	<< *current << endl;
#endif
	    
    everywhere = false;

    return(status);
}


/////////////////////////////////////////////////////////////////////
// @SUBTITLE "ActiveSet::which: ifp support"
//   See ActiveSet.h for #define ifp().
// We test and increment the whichLoop variable.
// If its old value was 0, we do THEN clause (return true)
// If its old value was 1, we do ELSE clause (return false)

boolean				// which clause? true for then, false for else
ActiveSet::which(void)
{
    return (whichLoop++ == 0);	// 0 for then, 1 for else
}
